/*
 * adau1361.c
 *
 * Copyright(C) 2014 Kaneko System Co., Ltd.
 *
 * Last Update: 2014/09/17
 *
 * []
 * ̃\tgEFA̓t[\tgEFAłBǂȂłRɉςĎgp邱Ƃł܂B
 * ̃\tgEFA͖ۏ؂łB̃\tgEFA̎gpɋNA[U܂͑O҂
 * QɊւA؂̐ӔC𕉂܂B
 *
 * [NOTE]
 * This program is free software, you can redistribute it and/or modify it.
 * There is no warranty for this free software.
 *
 */
#include <def21489.h>
#include <cdef21489.h>
#define SRUDEBUG  // Check SRU Routings for errors.
#include <SRU.h>
#include <interrupt.h>
#include <services/int/adi_int.h>
#include "codec.h"

/*
   Here is the mapping between the SPORTS and the ADC/DAC
   ADC -> DSP : SPORT0A : I2S Channel
   DSP -> DAC : SPORT1A : I2S Channel
*/

#define PCI    0x00080000
#define OFFSET 0x00080000

// SPORT rx DMA buffers
section("seg_dmda") static int sport0_rxbuf[NUM_RX_SLOTS][NUM_SAMPLES * 2];
// SPORT tx DMA buffers
section("seg_dmda") static int sport1_txbuf[NUM_TX_SLOTS][NUM_SAMPLES * 2];
static volatile int sport0_use_bufidx = NUM_RX_SLOTS - 1;

// SPORT tx DMA buffers
section("seg_dmda") static float sport_rxbuf[2][NUM_SAMPLES];
section("seg_dmda") static float sport_txbuf[2][NUM_SAMPLES];

// Set up the TCBs to rotate automatically
static int tcb_sport0_rxbuf[2][4] = {
    { 0, sizeof(sport0_rxbuf[0]), 1, 0 },
    { 0, sizeof(sport0_rxbuf[1]), 1, 0 } };

static int tcb_sport1_txbuf[2][4] = {
    { 0, sizeof(sport1_txbuf[0]), 1, 0 },
    { 0, sizeof(sport1_txbuf[1]), 1, 0 } };

/* f[^]tO */
static volatile int data_incoming = 0;

static const unsigned char audio_init_clock_off[] =
{
    // R0: Clock Control
    0x40, 0x00, 0x00,       // 0000-0000
};

static const unsigned char audio_init_clock_on[] =
{
    // R0: Clock Control
    0x40, 0x00, 0x0F,       // 0000-1111
};

static const unsigned char audio_init_pll[] =
{
    // R1: PLL
    0x40, 0x02, 0x00, 0x7D, 0x00, 0x0C, 0x21, 0x01, // 0x007D 000C 2101
};

/* ADAU1361 SPIf[^ */
static const unsigned char audio_init_data[] =
{
    // R2: Digital Microphone/Jack Detection Control
    0x40, 0x08, 0x00,
    // R3: Record Power Management
    0x40, 0x09, 0x74,       // 0111-0100
    // R4: Record Mixer Left (Mixer 1) Control 0
    0x40, 0x0A, 0x5B,       // 0101-1011
    // R5: Record Mixer Left (Mixer 1) Control 1
    0x40, 0x0B, 0x00,       // 0000-1000
    // R6: Record Mixer Right (Mixer 2) Control 0
    0x40, 0x0C, 0x5B,       // 0101-1011
    // R7: Record Mixer Right (Mixer 2) Control 1
    0x40, 0x0D, 0x00,       // 0000-0000
    // R8: Left Differential Input Volume Control
    0x40, 0x0E, 0x00,
    // R9: Right Differential Input Volume Control
    0x40, 0x0F, 0x00,
    // R10: Record Microphone Bias Control
    0x40, 0x10, 0x00,
    // R11: ALC Control 0
    0x40, 0x11, 0x03,       // 0000-0011
    // R12: ALC Control 1
    0x40, 0x12, 0x00,
    // R13: ALC Control 2
    0x40, 0x13, 0xF0,       // 0100-0000
    // R14: ALC Control 3
    0x40, 0x14, 0x64,       // 0110-0100

    // R15: Serial Port Control 0
    0x40, 0x15, 0x01,       // 0000-0001
    // R16: Serial Port Control 1
    0x40, 0x16, 0x00,
    // R17: Converter Control 0
    0x40, 0x17, 0x00, // 48kHz
//    0x40, 0x17, 0x06, // 96kHz
    // R18: Converter Control 1
    0x40, 0x18, 0x00,
    // R19: ADC Control
    0x40, 0x19, 0x23,       // 0010-0011
    // R20: Left Input Digital Volume
    0x40, 0x1A, 0x00,
    // R21: Right Input Digital Volume
    0x40, 0x1B, 0x00,
    // R22: Playback Mixer Left (Mixer 3) Control 0
    0x40, 0x1C, 0x21,       // 0010-0001
    // R23: Playback Mixer Left (Mixer 3) Control 1
    0x40, 0x1D, 0x00,
    // R24: Playback Mixer Right (Mixer 4) Control 0
    0x40, 0x1E, 0x41,       // 0100-0001
    // R25: Playback Mixer Right (Mixer 4) Control 1
    0x40, 0x1F, 0x00,
    // R26: Playback L/R Mixer Left (Mixer 5) Line Output Control
    0x40, 0x20, 0x05,       // 0000-0101
    // R27: Playback L/R Mixer Right (Mixer 6) Line Output Control
    0x40, 0x21, 0x11,       // 0001-0001
    // R28: Playback L/R Mixer Mono Output
    0x40, 0x22, 0x00,
    // R29: Playback Headphone Left Volume Control
    0x40, 0x23, 0xEF,       // 1110-1111
    // R30: Playback Headphone Right Volume Control
    0x40, 0x24, 0xEF,       // 1110-1111
    // R31: Playback Line Output Left Volume Control
    0x40, 0x25, 0xEE,       // 1110-1110
    // R32: Playback Line Output Right Volume Control
    0x40, 0x26, 0xEE,       // 1110-1110
    // R33: Playback Mono Output Control
    0x40, 0x27, 0x00,
    // R34: Playback Pop/Click Suppression
    0x40, 0x28, 0x00,
    // R35: Playback Power Management
    0x40, 0x29, 0xAB,       // 1010-1011
    // R36: DAC Control 0
    0x40, 0x2A, 0x03,       // 0000-0011
    // R37: DAC Control 1
    0x40, 0x2B, 0x00,
    // R38: DAC Control 2
    0x40, 0x2C, 0x00,

    // R39: Serial Port Pad Control
    0x40, 0x2D, 0xAA,       // 1010-1010
    // R40: Control Port Pad Control 0
    0x40, 0x2F, 0xAA,       // 1010-1010
    // R41: Control Port Pad Control 1
    0x40, 0x30, 0x00,
    // R42: Jack Detect Pin Control
    0x40, 0x31, 0x08,       // 0000-1000
    // R67: Dejitter Control
    0x40, 0x36, 0x03,
};

static void adau1361_sport_rx_isr(unsigned int iid, void* p_arg)
{
    /* Sampling data arrival*/
    data_incoming = 1;
    sport0_use_bufidx = (sport0_use_bufidx + 1U) % NUM_RX_SLOTS;
}

static void sharc_sport0_init(void)
{
    // Connect to the ADC
    SRU(DAI_PB18_O, SPORT0_CLK_I);
    SRU(DAI_PB17_O, SPORT0_FS_I);
    SRU(DAI_PB20_O, SPORT0_DA_I);

    SRU(LOW, PBEN18_I);
    SRU(LOW, PBEN17_I);
    SRU(LOW, PBEN20_I);

    // Unmask SPORT0 RX ISR Interrupt
    adi_int_InstallHandler(ADI_CID_P6I, adau1361_sport_rx_isr, 0, true);

    // Set up "ping-pong" chained DMAs for ADAU1361
    // Proceed from Block [0] to Block [1]
    tcb_sport0_rxbuf[0][0] = (int)&tcb_sport0_rxbuf[1][0] + 3 - OFFSET + PCI;
    tcb_sport0_rxbuf[0][3] = (int)&sport0_rxbuf[0][0] - OFFSET;

    // Proceed from Block [1] to Block [0]
    tcb_sport0_rxbuf[1][0] = (int)&tcb_sport0_rxbuf[0][0] + 3 - OFFSET + PCI;
    tcb_sport0_rxbuf[1][3] = (int)&sport0_rxbuf[1][0] - OFFSET;

    // Clear out SPORT0 registers
    *pSPMCTL0 = 0;
    *pSPCTL0 = 0;

    // Configure SPORT0 as a reciever
    // externally generating clock and frame sync
    *pDIV0 = 0;
    *pSPCTL0 = 
        SPEN_A |    // Enable Channel A
        L_FIRST |   // I2S sends left first
        SLEN32 |    // 32-bit word length
        OPMODE |    // Enable I2S Mode
        SDEN_A |    // Enable Channel A DMA
        SCHEN_A;    // Enable Channel A DMA Chaining
    // Enabling Chaining
    *pCPSP0A = tcb_sport0_rxbuf[1][0];
}

static void sharc_sport1_init(void)
{
    // Connect to the DAC
    SRU(DAI_PB18_O, SPORT1_CLK_I);
    SRU(DAI_PB17_O, SPORT1_FS_I);
    SRU(SPORT1_DA_O, DAI_PB19_I);

    SRU(HIGH, PBEN19_I);

    // Proceed from Block [0] to Block [1]
    tcb_sport1_txbuf[0][0] = (int)&tcb_sport1_txbuf[1][0] + 3 - OFFSET;
    tcb_sport1_txbuf[0][3] = (int)&sport1_txbuf[0][0] - OFFSET;

    // Proceed from Block [1] to Block [0]
    tcb_sport1_txbuf[1][0] = (int)&tcb_sport1_txbuf[0][0] + 3 - OFFSET;
    tcb_sport1_txbuf[1][3] = (int)&sport1_txbuf[1][0] - OFFSET;

    // Configure SPORT1 as a transmitter
    *pDIV1 = 0;
    *pSPCTL1 = 
        SPEN_A |    // Enable Channel A
        L_FIRST |   // I2S sends left first
        SLEN32 |    // 32-bit word length
        SPTRAN |    // TX mode
        OPMODE |    // Enable I2S Mode
        SDEN_A |    // Enable Channel A DMA
        SCHEN_A;    // Enable Channel A DMA Chaining
    // Enabling Chaining
    *pCPSP1A = tcb_sport1_txbuf[1][0];
}

//#define SELECT_SPI_SLAVE(select)   (*pSPIFLG &= ~(select << 8))
//#define DESELECT_SPI_SLAVE(select) (*pSPIFLG |= (select << 8))
#define SELECT_SPI_SLAVE(select)   SRU(LOW, DPI_PB07_I)
#define DESELECT_SPI_SLAVE(select) SRU(HIGH, DPI_PB07_I)
#define NOP asm volatile("nop;")
#define SPI_DELAY NOP;NOP;NOP;NOP;NOP;NOP;NOP;NOP;NOP;NOP

static void adau1361_reg_write(unsigned char addr0, unsigned char addr1, unsigned char regdata)
{
    SELECT_SPI_SLAVE(DS0EN);
    SPI_DELAY;

    while ((*pSPISTAT & TXS) != 0) continue;
    *pTXSPI = 0x00;
    while ((*pSPISTAT & TXS) != 0) continue;
    *pTXSPI = addr0;
    while ((*pSPISTAT & TXS) != 0) continue;
    *pTXSPI = addr1;
    while ((*pSPISTAT & TXS) != 0) continue;
    *pTXSPI = regdata;

    // Wait for the SPI to indicate that it has finished.
    while ((*pSPISTAT & SPIF) == 0) continue;
    while ((*pSPISTAT & SPIFE) == 0) continue;

    SPI_DELAY;
    DESELECT_SPI_SLAVE(DS0EN);
    SPI_DELAY;
}

static void adau1361_pllreg_write(unsigned char addr0, unsigned char addr1, const unsigned char regdata[])
{
    int l_lc;

    SELECT_SPI_SLAVE(DS0EN);
    SPI_DELAY;

    while ((*pSPISTAT & TXS) != 0) continue;
    *pTXSPI = 0x00;
    while ((*pSPISTAT & TXS) != 0) continue;
    *pTXSPI = addr0;
    while ((*pSPISTAT & TXS) != 0) continue;
    *pTXSPI = addr1;
    for(l_lc = 0; l_lc < 6; l_lc++)
    {
        while ((*pSPISTAT & TXS) != 0) continue;
        *pTXSPI = regdata[l_lc];
    }

    // Wait for the SPI to indicate that it has finished.
    while ((*pSPISTAT & SPIF) == 0) continue;
    while ((*pSPISTAT & SPIFE) == 0) continue;

    SPI_DELAY;
    DESELECT_SPI_SLAVE(DS0EN);
    SPI_DELAY;
}

static unsigned char adau1361_reg_read(unsigned char addr0, unsigned char addr1)
{
    unsigned char l_rv;

    SELECT_SPI_SLAVE(DS0EN);
    SPI_DELAY;

    while ((*pSPISTAT & TXS) != 0) continue;
    *pTXSPI = 0x01;
    while ((*pSPISTAT & TXS) != 0) continue;
    *pTXSPI = addr0;
    while ((*pSPISTAT & TXS) != 0) continue;
    *pTXSPI = addr1;
    while ((*pSPISTAT & TXS) != 0) continue;
    *pTXSPI = 0xFF;

    // Wait for the SPI to indicate that it has finished.
    while ((*pSPISTAT & SPIF) == 0) continue;
    while ((*pSPISTAT & SPIFE) == 0) continue;
    l_rv = *pRXSPI;

    SPI_DELAY;
    DESELECT_SPI_SLAVE(DS0EN);
    SPI_DELAY;

    return l_rv;
}

static void sharc_spi_init(void)
{
//    SRU(SPI_FLG0_O, DPI_PB07_I);     // Connect SPI FLAG0 to DPI PB7.
    SRU(SPI_CLK_O,  DPI_PB08_I);     // Connect SPI CLK to DPI PB8.
    SRU(DPI_PB09_O, SPI_MISO_I);     // Connect DPI PB9 to MISO.
    SRU(SPI_MOSI_O, DPI_PB10_I);     // Connect MOSI to DPI PB10.

//    SRU(SPI_FLG0_PBEN_O, DPI_PBEN07_I);
    SRU(SPI_CLK_PBEN_O,  DPI_PBEN08_I);
    SRU(SPI_MISO_PBEN_O, DPI_PBEN09_I);
    SRU(SPI_MOSI_PBEN_O, DPI_PBEN10_I);

    SRU(HIGH, DPI_PB07_I);
    SRU(HIGH, DPI_PBEN07_I);

    // Configure the SPI Control registers
    // First clear a few registers
    *pSPICTL = TXFLSH | RXFLSH;
    *pSPIFLG = 0;

    // Setup the baud rate to 1MHz
    *pSPIBAUD = 100;

    // Setup the SPI Flag register using the Flag specified in the call
    *pSPIFLG = SPIFLG0 | SPIFLG1 | SPIFLG2 | SPIFLG3 | DS0EN;

    // Now setup the SPI Control register
    *pSPICTL = SPIEN | SPIMS | WL8 | MSBF | TIMOD1 | SMLS | GM;
}

void adau1361_init(void)
{
    int l_lc;
    int l_ld;

    sharc_spi_init();

    /* Enter SPI mode */
    for(l_lc = 0; l_lc < 3; l_lc++)
    {
        SELECT_SPI_SLAVE(DS0EN);
        for(l_ld = 0; l_ld < 100000; l_ld++) NOP;
        DESELECT_SPI_SLAVE(DS0EN);
        for(l_ld = 0; l_ld < 100000; l_ld++) NOP;
    }

    /* Initialze ADAU1361 */
    adau1361_reg_write(audio_init_clock_off[0], audio_init_clock_off[1], audio_init_clock_off[2]);
    adau1361_pllreg_write(audio_init_pll[0], audio_init_pll[1], &audio_init_pll[2]);
    for(l_ld = 0; l_ld < 100000; l_ld++) NOP;
    adau1361_reg_write(audio_init_clock_on[0], audio_init_clock_on[1], audio_init_clock_on[2]);

    for(l_lc = 0; l_lc < sizeof(audio_init_data); l_lc += 3)
    {
        while(1)
        {
            adau1361_reg_write(audio_init_data[l_lc + 0], audio_init_data[l_lc + 1], audio_init_data[l_lc + 2]);
            if(adau1361_reg_read(audio_init_data[l_lc + 0], audio_init_data[l_lc + 1]) == audio_init_data[l_lc + 2]) break;
        }
    }

    sharc_sport0_init();
    sharc_sport1_init();
}

int adau1361_data_arrival_test(void)
{
    int l_dmabuf_idx;
    int l_lc, l_ld;
    int l_id;
    float l_fd;

    if(data_incoming)
    {
        // Callback audio processing function
        l_dmabuf_idx = sport0_use_bufidx;

        for(l_lc = 0, l_ld = 0; l_ld < NUM_SAMPLES; l_lc += 2, l_ld++)
        {
            l_id = sport0_rxbuf[l_dmabuf_idx][l_lc + 0];
            asm("%0 = float %1 by %2;" : "=F"(l_fd) : "d"(l_id), "d"(-31));
            sport_rxbuf[0][l_ld] = l_fd;

            l_id = sport0_rxbuf[l_dmabuf_idx][l_lc + 1];
            asm("%0 = float %1 by %2;" : "=F"(l_fd) : "d"(l_id), "d"(-31));
            sport_rxbuf[1][l_ld] = l_fd;
        }

        audio_process(sport_rxbuf, sport_txbuf);

        for(l_lc = 0, l_ld = 0; l_ld < NUM_SAMPLES; l_lc += 2, l_ld++)
        {
            l_fd = sport_txbuf[0][l_ld];
            asm("%0 = fix %1 by %2;" : "=d"(l_id) : "F"(l_fd), "d"(31));
            sport1_txbuf[l_dmabuf_idx][l_lc + 0] = l_id;

            l_fd = sport_txbuf[1][l_ld];
            asm("%0 = fix %1 by %2;" : "=d"(l_id) : "F"(l_fd), "d"(31));
            sport1_txbuf[l_dmabuf_idx][l_lc + 1] = l_id;
        }

        // Wait for the DMA completion
        data_incoming = 0;
        return 1;
    }
    else
    {
        return 0;
    }
}
