/*
 * sharc_pll.c
 *
 * Copyright(C) 2014 Kaneko System Co., Ltd.
 *
 * Last Update: 2014/09/17
 *
 * []
 * ̃\tgEFA̓t[\tgEFAłBǂȂłRɉςĎgp邱Ƃł܂B
 * ̃\tgEFA͖ۏ؂łB̃\tgEFA̎gpɋNA[U܂͑O҂
 * QɊւA؂̐ӔC𕉂܂B
 *
 * [NOTE]
 * This program is free software, you can redistribute it and/or modify it.
 * There is no warranty for this free software.
 *
 */
#include <def21489.h>
#include <cdef21489.h>

#define NOP asm volatile("nop;")

#define SDRAM_ADDR ((unsigned int*)0x200000)

static void sharc_pll_sdram_init(void)
{
    *pSYSCTL |= MSEN; // Enables FLAG2 and 3 (IRQ2 and TIMEXP) as MS2 and 3

     // Mapping Bank 0 to SDRAM
    *pEPCTL |=B0SD;
    *pEPCTL &= ~(B1SD | B2SD | B3SD);

    // Programming the SDCTL register
    *pSDCTL= SDCL3 | SDPSS | SDCAW9 | SDRAW13 | SDTRAS7 | SDTRP3 | SDTWR2 | SDTRCD3 | X16DE;

    // Change this value to optimize the performance for quazi-sequential accesses (step > 1)
    #define SDMODIFY 1                    // Setting the Modify to 1

    *pSDRRC = (0x4D8) | (SDMODIFY << 17) | SDROPT;

    *pSYSCTL |= EPDATA32; // Data [31:0] pins are configured as data lines

}

void sharc_pll_init(void)
{
    int l_lc;
    int l_reg;

    // Set INDIV bit in PMCTL register
    l_reg = *pPMCTL;
    l_reg |= INDIV;
    *pPMCTL= l_reg;

    // Program PLL multiplier to same value as CLK_CFGx pins/previously programmed value in software
    *pPMCTL = l_reg;

    // then place PLL in bypass mode
    l_reg |= PLLBP;
    *pPMCTL = l_reg;

    // Wait for recommended number of cycles
    for (l_lc = 0; l_lc < 4096; l_lc++)
        NOP;

    // Bring PLL out of bypass mode by clearing PLLBP bit
    *pPMCTL ^= PLLBP;

    for (l_lc = 0; l_lc < 16; l_lc++)
        NOP;

    l_reg = *pPMCTL;
    // Clear the previous PLL multiplier
    l_reg &= ~PLLM63;
    // Clear the INDIV bit
    l_reg &= ~INDIV;
    // or set the INDIV bit
    // *pMCTL |= INDIV;
    *pPMCTL= l_reg;

    l_reg = PLLM16 | PLLD2 | SDCKR2_5 | DIVEN;
    *pPMCTL = l_reg;

    l_reg |= PLLBP;    // Setting the Bypass bit
    l_reg ^= DIVEN;    // Clearing the DIVEN bit
    *pPMCTL = l_reg;   // Putting the PLL into bypass mode

    // Wait for around 4096 cycles for the pll to lock.
    for (l_lc = 0; l_lc < 5000; l_lc++)
        NOP;

    l_reg = *pPMCTL;
    l_reg ^= PLLBP;          // Clear Bypass Mode
    *pPMCTL = l_reg;

    // Wait for around 15 cycles for the output dividers to stabilize.
    for(l_lc = 0; l_lc < 16; l_lc++)
        NOP;

    sharc_pll_sdram_init();
}
